#! /usr/bin/env python

import numpy
from pbio import PbWriter

# Make some data - four variables defined on a sphere

attrnames = ['lon', 'lat', 'simplefunc', 'wildfunc']

# longitudes, from 0 to 2pi
lons = numpy.linspace( 0, 2*numpy.pi, num=120, endpoint=False )

# latitudes, from -pi/2 to pi/2
lats = numpy.linspace( -0.5*numpy.pi, 0.5*numpy.pi, num=60, endpoint=True )

# make space for grid of points and attributes
npts = len(lons) * len(lats)
nattrs = len(attrnames)

xyz = numpy.zeros( shape=(npts, 3) )		      # particle positions
ids = numpy.zeros( shape=(npts,), dtype=numpy.int )   # integer particle ids, even though we don't need them here
attrs = numpy.zeros( shape=(npts, nattrs) )	      # particle attributes


# There are much better ways to fill in xyz[] and attrs[], but this is simple & clear
i = 0
for lat in lats:
    for lon in lons:
	xyz[i,0] = numpy.cos(lon) * numpy.cos(lat)
	xyz[i,1] = numpy.sin(lon) * numpy.cos(lat)
	xyz[i,2] = numpy.sin(lat)
	attrs[i,0] = lon
	attrs[i,1] = lat
	attrs[i,2] = numpy.cos( 3 * lon )	# simple function
	attrs[i,3] = numpy.cos( 3 * lon + 3 * numpy.tan( 0.9*lat ) )  # wilder function
	ids[i] = i
	i += 1

pbwriter = PbWriter( 'sphere-example.pb', attrnames )
pbwriter.writepcles( xyz, attrs, id=ids )
pbwriter.close()

print "#! /usr/bin/env partiview"
print "# Paste this into a script, e.g. sphere-example2.cf, and run: partiview sphere-example2.cf"
print "pb sphere-example.pb # read the data"
print "eval lum const 5 # uniform brightness for all particles"
print "eval color wildfunc -1.01 1.01 # slightly bigger data range than wildfunc's actual values"
print "# \"color wildfunc\" or \"color simplefunc\" or \"color lon\" or \"color lat\""
print "eval jump -2.5 -5.3 3.5  51 -36 -38"

