#! /usr/bin/perl


sub acos {
    57.296 * atan2( sqrt(1-$_[0]**2), $_[0] );
}

$MAGSCALE = log( 100 ) / 5;
$rad = 3.1415926/180;
$arcsec = 1/3600 * $rad;
$LOG10 = log(10);

$H0 = 70;

$G_AGNGALAXY = 0;	# Galaxies-with-AGNs from sdss+gals+agn.csv)

$G_TYPEI_CORE = 1;	# Type-I quasar group (from sdss.csv) spikes
$G_TYPEI_HALO = 4;	# Type-I quasar group (from sdss.csv) haloes

$G_TYPEII_CORE = 2;	# Type-II quasar group (from sdss+gals+agn.csv)
$G_TYPEII_HALO = 6;	# Type-II quasar group (ditto) (haloes)

$G_GALAXY = 5;		# run-of-the-mill galaxies (from sdss-gals-again.csv)

$CLOUD = 1;
$SPIKE = 4;

# z of .01 corresp 3000 km/sec RV; take H_0 of 70 km/sec/Mpc,
#   so for small z, distance =~ z * c / H_0
# 1 Lsun has mag ~+5 at 10pc, mag +30 at 1Mpc,
# or 1e8 Lsun has mag +10 at 1Mpc.

sub unit {
    my($x,$y,$z) = @_;
    my($s) = sqrt( $x*$x + $y*$y + $z*$z );
    $s = 1 if $s == 0;
    ($x/$s, $y/$s, $z/$s);
}

sub remap {  # in, in0, in1, outrange, outrange => out
    my($in, $in0, $in1, $outrange, $out0) = @_;
    $outrange = 1 unless defined($outrange);
    my($o) = ($in - $in0) / ($in1 - $in0);
    $o = 0 if $o < 0;
    $o = 1 if $o > 1;
    $out0 + $outrange * $o;
}

sub putspeck {
    my(@p) = splice(@_, 0, 3);
    my($lum, $cindex, $radius, $group, $agnfrac) = @_;

    printf defined($agnfrac) ? "%g %g %g  %g %d %g  %g %g\n"
			     : "%g %g %g  %g %d %g  %g\n",
	@p, $lum, $cindex, $radius, $group,
		$agnfrac;
}

sub zradec2distxyz {
    my($z, $ra, $dec) = @_;

    my($distmpc) = $z * 3.0e5 / $H0;
    my($sdec)    = sin($dec*$rad);
    my($cdec)    = cos($dec*$rad);
    my($sra)     = sin($ra*$rad);
    my($cra)     = cos($ra*$rad);

    ( $distmpc, $distmpc*$cra*$cdec, $distmpc*$sra*$cdec, $distmpc*$sdec );
}

sub putobject {
    my($name, $fields) = @_;

    print "#! /usr/bin/env partiview\n" unless $objectno;
    ++$objectno;

    print "\n";
    print "object g$objectno=$name\n";

print <<EOF;

datavar 0 lum
datavar 1 color
datavar 2 radius
datavar 3 group
EOF

    print "datavar 4 agnfrac # OIII / Hbeta\n" if $fields >= 4;
    print "\n";
}

&putobject( "agnhostgals", 3 );

# Cook up fake 3-D spatial parameters for quasars, based on:
#   from gals+sdss+agn.csv: these are type-II quasars (seen off-axis)
# oid,z,ra,dec,mag_g,mag_r,mag_i,emag_r,AB,phi,rad_r,rad_i,plate,mjd,fiberid,modelMag_g,modelMag_r,modelMag_i,modelMag_z,o3corr,bpt1,vdisp

open(SGA, "gals+sdss+agn.csv") or print STDERR "gals+sdss+agn.csv: $!";

$_ = <SGA>;
@fields = split(' ');

$o3lmax = 0;

while(<SGA>) {
    ($oid,$z,$ra,$dec,$mag_r,$mag_i,$AB,$rad_r,$modelMag_r,$modelMag_i,$o3corr, $bpt1,$vdisp) =
	( split(/,/) )
    [ 0,   1,  2,  3,    5,     6,   8,   10,      16,         17,        19,      20,   21 ];
# oid,z,ra,dec,mag_g,mag_r,mag_i,emag_r,AB,phi,rad_r,rad_i,plate,mjd,fiberid,modelMag_g,modelMag_r,modelMag_i,modelMag_z,o3corr,bpt1,vdisp
#  0  1  2  3    4     5     6     7    8   9   10    11     12  13    14       15         16          17        18        19   20     21

    next if $o3corr > 20 or $z <= 0;	# unusable

    next if $modelMag_r eq "-9999" or $modelMag_i eq "-9999" or $mag_r eq "-9999" or $mag_i eq "-9999" or $AB eq "-9999";

    $seenoid{$oid} = 1;

    my( $distmpc, @p ) = &zradec2distxyz( $z, $ra, $dec );

    $lsun_r = exp( (30 - $mag_r) * $MAGSCALE ) * $distmpc*$distmpc;
    $o3lum = exp( $o3corr * $LOG10 );
    $o3_hbeta = $bpt1 > 5 ? 0 : exp( $bpt1 * $LOG10 );

    if($o3lmax < $o3lum || $o3lum > $lsun_r) {
	printf STDERR "o3lum %g Lsun_r %g o3/L %g o3/Hbeta %g mag_r %.2f z %.4f\n", $o3lum, $lsun_r, $o3lum/$lsun_r, $o3_hbeta, $mag_r, $z;
	$o3lmax = $o3lum;
    } 

    $radmpc = $distmpc * $rad_r * $arcsec;

    $num = $oid % 1000000000;

    # These are type-II quasars.  Emit two particles:
    #    group 0 bright_cloud with color[0..1023]+lum+diam of host galaxy,
    #    group 2 spike             color[2048..3071] from o3lum/lsun_r, lum = o3lum, d{xyz} random vector

    $cindex0 = &remap( $mag_r - $mag_i, -1, 1, 1024, 0 );
    &putspeck( @p, $lsun_r, $cindex0, $radmpc, $G_AGNGALAXY, $o3_hbeta );  # pos, lum, cindex, radius, type, group

    ## construct a deterministically "random" unit vector
    #srand( $oid );
    #@axis = &unit( rand(2)-1, rand(2)-1, rand(2)-1 );

    # color directly by vdisp -- 1 km/sec per cindex slot.  Most are 0..600.
    # This is colortable group 2, so cindex range 2048 .. 2048+1023
    
    #$cindex2 = &remap( $vdisp, 0, 1023,  1023, 2048 );
    $cindex2 = &remap( $o3lum/$lsun_r, 0, 1,  1023, 2048 );
    push( @agns, $o3lum, $cindex2, $o3_hbeta, @p ); 
    $quagals++;
}
close(SGA);

&putobject( "agns", 4 );

print "eval polysize 1\n";
print "eval polylumvar radius\n";
print "eval txno radius\n";
print "eval polysides 4\n";
print "eval texture -M 1 ring.pgm\n";
print "\n";

while(@agns) {
    my( $o3lum, $cindex2, $o3_hbeta, @p ) = splice( @agns, 0, 6 );
    &putspeck( @p, $o3lum, $cindex2, 1, $G_TYPEII_CORE, $o3_hbeta );
}
print "\n";


&putobject( "quasars", 3 );


open(SDSSQ, "sdss.csv") or print STDERR "sdss.csv: $!\n";
$_ = <SDSSQ>;
while(<SDSSQ>) {
    ($plate,$mjd,$fiberid,$specClass,$z,$ra,$dec,$mag_r,$mag_i) =
	( split(/,/) )
    [   0,    1,    2,        3,      4,  5,  6,    9,    10  ];
#  plate,mjd,fiberid,specClass,z,ra,dec,modelMag_u,modelMag_g,modelMag_r,modelMag_i,modelMag_z
#    0    1     2        3     4  5  6     7           8          9         10         11

    next unless $specClass == 3 && $z > 0 && $z < 0.4 && ($mag_r > 10 && $mag_r < 30);

    my( $distmpc, @p ) = &zradec2distxyz( $z, $ra, $dec );

    $lsun_r = exp( (30 - $mag_r) * $MAGSCALE ) * $distmpc*$distmpc;

    $num = substr( $plate . $mjd . $fiberid, -9 );

    ## Construct axis for this quasar.  Pick something within about 20 degrees of
    ## the quasar<->Earth direction.
    #
    #srand( $plate . $mjd . $fiberid );
    #@pu = &unit( @p );
    #$pu[0] += 0.3 * (rand(2) - 1);
    #$pu[1] += 0.3 * (rand(2) - 1);
    #$pu[2] += 0.3 * (rand(2) - 1);
    #@axis = &unit( @pu );

    $cindex1 = &remap( $mag_r - $mag_i, -1, 1, 1023, 1024 );
    &putspeck( @p, $lsun_r, $cindex1, 0, $G_TYPEI_HALO );
    $quasars++;
}
close(SDSSQ);


&putobject( "gals" );

open(JUSTGALS, "../sdss-gals-again.csv") or print STDERR "../sdss-gals-again.csv: $!\n";
$_ = <JUSTGALS>;
while(<JUSTGALS>) {
    my($oid,$specClass,$z,$ra,$dec,$mag_r,$mag_i,$AB,$phi,$rad_r,$conc90_50) =
	( split(/,/) )
      [  0,     2,      1,  3,  4,    6,     7,   9,  10,   11,     13     ];
#oid,z,specClass,ra,dec,mag_g,mag_r,mag_i,emag_r,AB,phi,rad_r,rad_i,conc90_50
# 0  1     2      3  4   5      6     7     8     9  10  11    12       13

    next unless $specClass == 2 and $z > 0 and $mag_r > 10 and $mag_r < 30 and $mag_i > 10 and $mag_i < 30;
    if(defined($seenoid{$oid})) {
	$dups++, next if $seenoid{$oid} == 1;
	$dups2++, next;
    }
    $seenoid{$oid} = 2;
    $num = substr( $oid, -9 );

    my( $distmpc, @p ) = &zradec2distxyz( $z, $ra, $dec );
    
    $lsun_r = exp( (30 - $mag_r) * $MAGSCALE ) * $distmpc*$distmpc;

    $radmpc = $distmpc * $rad_r * $arcsec;

    # These are plain galaxies.  Emit one particle:
    #    group 0 bright_cloud with color[0..1023]+lum+diam of host galaxy,
    #    group 2 spike             color[2048..3071] from vdisp, lum = o3lum, d{xyz} random vector

    $cindex0 = &remap( $mag_r - $mag_i, -1, 1, 1024, 0 );
    &putspeck( @p, $lsun_r, $cindex0, $radmpc, $G_GALAXY );  # pos, lum, cindex, radius, OIII/Hbeta, group
    $galsagain++;
}
print STDERR "$galsagain galaxies (skipping $dups/$dups2 dups), $quasars type I, $quagals type II quasars\n";

#oid,z,ra,dec,mag_g,mag_r,mag_i,emag_r,AB,phi,rad_r,rad_i,plate,mjd,fiberid,modelMag_g,modelMag_r,modelMag_i,modelMag_z,o3corr,bpt1,vdisp
#587722981733564567,0.07434339821338653,165.08793,-1.1253654,19.140300750732422,18.25790023803711,17.87350082397461,16.886072158813476,0.44569945335388183,149.369140625,2.679915428161621,2.6630756855010986,277,51908,210,17.38483,16.57016,16.15084,15.83376,6.5541,-0.1203,100.5116
    

# for gals+sdss+agn.csv: (Garching catalog entries with galaxy profiles)
#   AB -> inclination
#   z, mag_r -> apparent M_r
#   M_r, inclination, beamwidth -> intrinsic M_r

# for sdss.csv where specClass=3 (QSO; no galaxy profile available):
#   assume small random inclination
#   z, mag_r -> apparent M_r
#   M_r, inclination, beamwidth -> intrinsic M_r
